#!/usr/bin/perl -w
#
# Multi-user Irc Relay Daemon
#
# finished Friday April 14 12:45 AM
# see pod at end of file, or 'perldoc ./mird.pl'
#
# edit these defaults

my $def_port = 3456;                      # local port
my $def_maxcli = 10;                      # max num of clients
my $def_firstserver = "irc.nethead.com";  # default initial connection

use Getopt::Std;

my %opts;

getopts('fm:p:s:',\%opts);

my $port;
if ($port = $opts{'p'}) {
    print "Setting port to $port\n"; 
} else {
    print "Setting port to default value $def_port\n";
    $port = $def_port;
}

my $maxcli;
if ($maxcli = $opts{'m'}) {
    print "Setting max clients to $maxcli\n"; 
} else {
    print "Setting max clients to default value $def_maxcli\n";
    $maxcli = $def_maxcli;
}

my $def_server;
if ($def_server = $opts{'s'}) {
    print "Setting first server to $def_server\n"; 
} else {
    print "Setting first server to default value $def_firstserver\n";
    $def_server = $def_firstserver;
}

use IO::Select;
use IO::Socket;
use strict;

my $server = new IO::Socket::INET(
				  LocalPort => $port,
				  Proto => 'tcp',
				  Listen => $maxcli,
				  Reuse => 1
				  );

$server or die "Fatal server error: $@";

$opts{'f'} or daemonize();

my $selhan = new IO::Select($server);
my $irc_con = 0;

# structure for handling client and server data
my %i_stat = (
	      -sockstat => {                  # server stuff
		  -defserver => $def_server,  # default irc server
		  -server => $server,         # listen server
		  -selhan => $selhan,         # select handle
		  -irc_con_ref => \$irc_con,  # irc connection
		  -regged => 0,	              # registered flag
	      },
	      -channels => { },    
# channel info
	      -me => '',           # current nick
	      -server => '',       # current server name
	      -init_nick => '',    # initial nick connect string
	      -init_user => '',    # initial username
	      -clients => { },     # currently connected clients
	      -introtext => [ ],   # introductory text
	      );
	    
my $old_buf = ''; # buffer for irc connection
my %cli_buf = (); # buffer for client connection

while(1) {       
    foreach my $in_client ($selhan->can_read) {  # incoming data
	if ($in_client eq $server) { # new connection
	    if ((keys %{$i_stat{-clients}}) >= $maxcli) {
		my $tmp = $server->accept;
		print $tmp
		    "ERROR :Closing Link: Server is full\n";
		select(undef,undef,undef,".05");
		next;
	    }

	    my $new_client = $server->accept;
	    $cli_buf{$new_client} = '';
	    addClient(\%i_stat,$new_client); # add it to the list
	} elsif ($in_client eq $irc_con) { # irc data
	    my $data;
	    my $read = sysread($irc_con,$data,64000); # some big number
	    unless(defined $read and length $data) {
		$selhan->remove($irc_con);
		massChanNotice(\%i_stat, "Main irc connection down! " . 
			       "Someone type " .
			       "/quote SERVER servername or /raw SERVER " .
			       "servernams ASAP!!");
		# remove irc status info
		$i_stat{-channels} = { };
		${$i_stat{-sockstat}{-irc_con_ref}} = 0;
		$i_stat{-me} = '';
		$i_stat{-introtext} = [ ];
		$old_buf = '';
		next;
	    }	    
	    $data = $old_buf . $data; # buffering system for irc connection
	    $old_buf = '';
	    while($data =~ s/^(.*?)\n//s) {
		processIrcText(\%i_stat,$1);
	    }
	    $old_buf = $data;
	} else {
	    my $data;
	    my $read = sysread($in_client,$data,64000); # some big number
	    unless(defined $read and length $data) {
		delClient(\%i_stat,$in_client);
		$selhan->remove($in_client);
		next;
	    }
	    $data = $cli_buf{$in_client} . $data; 
	    $cli_buf{$in_client} = '';
	    my $i = 1;
	    while($data =~ s/^(.+?)[\015\012]+//s) {
		processUserText(\%i_stat,$in_client,$1);
	    }
	    $cli_buf{$in_client} = $data;
	}
    }
}

# processess incoming text from client
#
sub processUserText {
    my($i_stat,$in_client,$data) = @_;

    my($sel,$iref) = ($i_stat->{-sockstat}{-selhan},
		      $i_stat->{-sockstat}{-irc_con_ref} );
    
    chomp($data);
    my(@parms) = split(' ',$data);
    
    if (uc($parms[0]) eq "SERVER") { # connect to irc server
	my($address,$port) = @parms[1,2];
	$port ||= 6667;
	unless($address) {
	    sendSock($i_stat,$in_client,"You didn't specify a server!\n");
	    return;
	}
	if ($$iref) {
	    $sel->remove($$iref);
	    close($$iref);
	    # print "Shutting down old connection\n";
	    massChanNotice($i_stat,"Hopping to $address:$port");
	    massTransmit($i_stat,
			 "ERROR :Closing Link: Hopping to new server!\n"
			 );

	} else {
	    massChanNotice($i_stat,"Logging on to $address:$port");
	}
	if (conIrc($iref,$address,$port)) { # attempt connection
	    massChanNotice($i_stat,"Connected!");
	    $sel->add($$iref);
	    my $tmp;
	    if ($tmp = $i_stat->{-init_nick}) {
		# print "Sent initial nick - $tmp\n";
		$$iref->print("$tmp\n");
	    }
	    if ($tmp = $i_stat->{-init_user}) {
		# print "Sent initial user - $tmp\n";
		$$iref->print("$tmp\n");
	    } 
	} else {
	    massChanNotice($i_stat,"Unable to connect!");
	    return;
	}
	return;
    } elsif (uc($parms[0]) eq "BIGNICK") { # change the main con's nick
	$$iref->print("NICK $parms[1]\n");
	return;
    } elsif (uc($parms[0]) eq "STATUS") { # shut it all down
	my @status;
	push(@status,"Server status: " . localtime() . " -");
	foreach my $client (keys %{$i_stat->{-clients}}) {
	    push(@status,
		 $i_stat->{-clients}{$client}{-nick} . " (" .
		 $i_stat->{-clients}{$client}{-host} . ")" );
	}
	foreach my $line (@status) { 
        # send status report from self - let client deal
	   print $in_client "$i_stat->{-me}!bogus\@bogus NOTICE " . 
	       "$i_stat->{-me} :$line\n";
	}
	return;
    } elsif (uc($parms[0]) eq "SHUTDOWN") { # shut it all down
	exit; # this could be more graceful
    } elsif (uc($parms[0]) eq "NICK") { # local nick change
	# flag the that client has requested a nick change
	my $newnick = $parms[1];
	$newnick =~ s/^://;
	my $oldnick = $i_stat->{-clients}{$in_client}{-nick};
	
	$i_stat->{-clients}{$in_client}{-nick} = $newnick;

	# if we are not connected, we want to save this nick string for
	#  when we do connect to an irc server
	unless($$iref) {
	    $i_stat->{-init_nick} = $data;

	    # do we have a complete USER and NICK set?
	    if (
		(scalar(keys(%{$i_stat->{-clients}})) == 1) and
		($i_stat->{-init_user})
		) {			
		if(conIrc($iref,$i_stat->{-sockstat}{-defserver})) { 
		    # transmit user and nick
		    $i_stat->{-sockstat}{-selhan}->add($$iref);
		    $$iref->print("$i_stat->{-init_user}\015\012"); 
		    $$iref->print("$i_stat->{-init_nick}\015\012");	   
		}
	    }		
	    return;
	}

	# we are connected, but not registered - just send in the nick now
	#  this may happen multiple times, but its ok
	unless ($i_stat->{-sockstat}{-regged}) { # we are not connected!
	    $$iref->print("$data\n");
	    return;
	}

	# check to see if this nick has been initialized, and initialize
	#  it if necessary - otherwise, transmit the nick change to other
	#  clients
	if ($i_stat->{-clients}{$in_client}{-init}) {
	    massChanNotice($i_stat,"$oldnick is now known as $newnick");
	} else {
	    initClient($i_stat,$in_client);
	}

	return;
    } elsif (uc($parms[0]) eq "USER") { # pretty much same as NICK
	unless($$iref) {
	    $i_stat->{-init_user} = $data;
	    if (
		(scalar(keys(%{$i_stat->{-clients}})) == 1) and
		($i_stat->{-init_nick})
		) {
		if(conIrc($iref,$i_stat->{-sockstat}{-defserver})) {
		    $i_stat->{-sockstat}{-selhan}->add($$iref);
		    $$iref->print("$i_stat->{-init_nick}\n");
		    $$iref->print("$i_stat->{-init_user}\n");
		}
	    }		
	    return;
	}

	unless ($i_stat->{-sockstat}{-regged}) { # we are not regged
	    $data =~ s/\015|\012//g;
	    $$iref->print("$data\n");
	}
	return;
    } elsif (uc($parms[0]) eq "QUIT") { # drop client and continue
	delClient($i_stat,$in_client);
	return; # do not transmit
    } elsif (
	     (uc($parms[0]) eq "MODE") and
	     (! defined($parms[2])) and
	     (exists $i_stat->{-channels}{lc($parms[1])}{-mode})
	     ) { # MODE request
		
	my $chan = lc($parms[1]);
	my $mode = $i_stat->{-channels}{$chan}{-mode};
	$i_stat->{-channels}{$chan}{-limit} and 
	    $mode .= $i_stat->{-channels}{$chan}{-limit};
	$i_stat->{-channels}{$chan}{-key} and 
	    $mode .= $i_stat->{-channels}{$chan}{-key};
	print $in_client ":$i_stat->{-server} 324 $i_stat->{-me} " . 
	    "$chan $mode\n";
	return;
    } elsif (uc($parms[0]) =~ /WHO\b|ISON|PING/) { # obnoxius client stuff
	return; # do not transmit
    } elsif (uc($parms[0]) eq "PRIVMSG") { # needs handling
	# send to local clients
	my($target,$text) = ($parms[1],$data);
	$text =~ s/^.*?://g;

	my $fromnick = $i_stat->{-clients}{$in_client}{-nick};

	# transmit privmsg back to other clients
	if ($target =~ /^(\#|\+|&)/) {   # channel
	    my $send = ":$i_stat->{-me}!bogus\@bogus PRIVMSG $target :" .
		"<$fromnick> $text";	
	    massTransmit($i_stat,"$send\n",$in_client);
	} else { # private
	    my $send = ":$target!bogus\@bogus PRIVMSG $target :" .
		"<$fromnick> $text";	
	    massTransmit($i_stat,"$send\n",$in_client);
	}

	unless ($text =~ /^\*/) { # local text
	    $$iref->print("$data\n");
	}
	return;
    }

#   print "From user: $data\n";

    unless($$iref) { # we aren't connected
	return;
    }    

    $$iref->print("$data\n");
}

# processes incoming text for irc
#
sub processIrcText {
    my($i_stat, $data) = @_;

#    print "Raw: $data\n";

    my $iref = $i_stat->{-sockstat}{-irc_con_ref};

    chomp($data);

    my($ping,$codes,$message) = split(':',$data,3);

    if ($ping =~ /PING/i)  {
	my $iref = $i_stat->{-sockstat}{-irc_con_ref};
	$codes =~ s/\015|\012//g;
	$$iref->print("PONG :$codes\n");
	return;
    }


    $message and my(@mparms) = split(' ',$message);
    my(@cpms) = split(' ',$codes);
    my($inick,$iident,$ihost) =	($cpms[0] =~ /(\S+)!(\S+)@(\S+)/);

# initial server messages

    if ($cpms[1] eq "001") { # this is who we are
	$i_stat->{-sockstat}{-regged} = 1; # registration complete
	$i_stat->{-me} = $cpms[2];
	$i_stat->{-server} = $cpms[0];
    } elsif ($cpms[1] =~ /002|003|004|005|375|272|376/) {
	push(@{$i_stat->{-introtext}},$data);  # intro stuff
    } 

# getting initial and paying attention to user/channel modes

    elsif ($cpms[1] eq "353") { # names responce from joining a chan
	while ($_ = shift(@mparms)) { # yes 0 is an erroneous nickname
	    my $mode = '';
	    s/\+// and $mode .= 'v';
	    s/\@// and $mode .= 'o';
	    # print "Adding $_ to chan $cpms[4]\n";
	    $i_stat->{-channels}{lc($cpms[4])}{$_} = "$mode";
	}
    } elsif ($cpms[1] eq "332") { # initial topic
	$i_stat->{-channels}{lc($cpms[3])}{-topic} = $message;
    } elsif ($cpms[1] eq "TOPIC") { # topic change
	$i_stat->{-channels}{lc($cpms[2])}{-topic} = $message;
    } elsif ($cpms[1] eq "324") { # initial mode
	$i_stat->{-channels}{lc($cpms[3])}{-mode} = $cpms[4];
    } elsif ($cpms[1] eq "329") { # channel created time - annoying!!
	return;
    } elsif (($cpms[1] eq "MODE") and ($cpms[2] =~ /^[\#+&]/)) {
	# channel mode change
        my @modes = split('',$cpms[3]);
        my $chan = lc($cpms[2]);
        splice(@cpms,0,4);

        my $curmode = "+";
        MINI: while(my $action = shift(@modes)) {
	  if ($action =~ /\+|\-/) {
	      $curmode = $action;
	      next;
	  }
	  if ($curmode eq "+") {
	      if ($action =~ /s|t|i|m|n|p/) {
		  $i_stat->{-channels}{$chan}{-mode} =~ s/$action//g;
		  $i_stat->{-channels}{$chan}{-mode} .= $action;
		  next MINI;
	      }
	      if ($action =~ /b|e/) {
		  shift(@cpms);
		  next MINI;
	      } elsif ($action eq "l") {
		  $i_stat->{-channels}{$chan}{-mode} =~ s/l//g;
		  $i_stat->{-channels}{$chan}{-mode} .= "l";
		  $i_stat->{-channels}{$chan}{-limit} =~ shift(@cpms);
		  next MINI;
	      } elsif ($action eq 'k') {
		  $i_stat->{-channels}{$chan}{-mode} =~ s/k//g;
		  $i_stat->{-channels}{$chan}{-mode} .= "k";
		  $i_stat->{-channels}{$chan}{-key} =~ shift(@cpms);
		  next MINI;
	      }

	      my $obj = shift(@cpms);
	      $i_stat->{-channels}{$chan}{$obj} =~ s/$action//g;
	      $i_stat->{-channels}{$chan}{$obj} .= $action;
	  } elsif ($curmode eq "-") {
	      if ($action =~ /s|t|i|m|n|p|l/) {
		  $i_stat->{-channels}{$chan}{-mode} =~ s/$action//g;
		  $action eq 'l' and 
		      delete $i_stat->{-channels}{$chan}{-limit};
		  next MINI;
	      }
	      elsif ($action =~ /b|e/) {
		  shift(@cpms);
		  next MINI;
	      } elsif ($action eq 'k') {
		  shift(@cpms);
		  delete $i_stat->{-channels}{$chan}{-key};
		  $i_stat->{-channels}{$chan}{-mode} =~ s/k//g;		  
		  next MINI;
	      }
	      my $obj = shift(@cpms);
	      $i_stat->{-channels}{$chan}{$obj} =~ s/$action//g;
	  }
      }

# all the ways to join and part a channel

    } elsif ($cpms[1] eq "JOIN") {
	$i_stat->{-channels}{lc($mparms[0])}{$inick} = '';
	$inick eq $i_stat->{-me} and 
	    delete $i_stat->{-channels}{lc($mparms[0])};
    } elsif ($cpms[1] eq "PART") {	
	delete $i_stat->{-channels}{lc($cpms[2])}{$inick};
	$inick eq $i_stat->{-me} and 
	    delete $i_stat->{-channels}{lc($cpms[2])};
    } elsif ($cpms[1] eq "KICK") {
	delete $i_stat->{-channels}{lc($cpms[2])}{$cpms[3]};
	$cpms[3] eq $i_stat->{-me} and 
	    delete $i_stat->{-channels}{lc($cpms[2])};
    }  elsif ($cpms[1] eq "QUIT") { # not channel specific
	foreach my $chan (keys %{$i_stat->{-channels}}) {
	    delete $i_stat->{-channels}{lc($chan)}{$inick};
	}
    } elsif ($cpms[1] eq "NICK") { # not channel specific
	if ($inick eq $i_stat->{-me}) {
	    $i_stat->{-me} = $mparms[0];
	}
	
	foreach my $chan (keys %{$i_stat->{-channels}}) {
	    if (exists($i_stat->{-channels}{lc($chan)}{$inick})) {
		my $tmpmode = $i_stat->{-channels}{lc($chan)}{$inick};
		delete $i_stat->{-channels}{lc($chan)}{$inick};
		$i_stat->{-channels}{lc($chan)}{$mparms[0]} = $tmpmode;
	    }	    
	}

    # have to handle CTCPs that generate responces seperately
	
    } elsif (($cpms[1] eq "PRIVMSG") and 
	     (substr($mparms[0],0,1) eq "\ca") and 
	     ($mparms[0] ne "\caDCC") and 
	     ($mparms[0] ne "\caACTION") 
	     ) {

	my $iref = $i_stat->{-sockstat}{-irc_con_ref};

	# I need more handlers here

	if ($mparms[0] eq "\caVERSION\ca") {
	    $$iref->print("NOTICE $inick :\caVERSION Go Away!\ca\n");
	    $cpms[2] eq $i_stat->{-me} ?
		massChanNotice($i_stat,"Detected CTCP VERSION from $inick!") :
		massChanNotice($i_stat,"Detect CTCP VERSION from $inick " .
			       "to $cpms[2]!");
	}
	return;
    }

    massTransmit($i_stat,"$data\n",undef);
}

# transmits raw text to every client (usually local /msg's),
#  skips client in $from
#
sub massTransmit {
    my($i_stat,$data,$from) = @_;
    foreach my $out_client (keys %{$i_stat->{-clients}}) {
        defined($from) and $out_client eq $from and next;
	my $sock = $i_stat->{-clients}{$out_client}{-socket};
	print $sock $data;
    }
}

# Sends a message to every client on every chan
#
sub massChanNotice {
    my($i_stat,$message,$skip) = @_;

    foreach my $out_client (keys %{$i_stat->{-clients}}) {
	defined($skip) and $out_client eq $skip and next;
	my $sock = $i_stat->{-clients}{$out_client}{-socket};
	print $sock ":$i_stat->{-me}!bogus\@bogus NOTICE $i_stat->{-me} :" .
	    "$message\n";
    }	

    # below is an older method, the above is much friendier to console clients

    return; 
    my $bool;
    foreach my $chan (keys %{$i_stat->{-channels}}) {
	$bool++;
	foreach my $out_client (keys %{$i_stat->{-clients}}) {
	    defined($skip) and $out_client eq $skip and next;
	    my $sock = $i_stat->{-clients}{$out_client}{-socket};
	    print $sock ":Server!bogus\@bogus NOTICE $chan :$message\n";
	}	
    }
    unless ($bool) {
	foreach my $out_client (keys %{$i_stat->{-clients}}) {
	    defined($skip) and $out_client eq $skip and next;
	    my $sock = $i_stat->{-clients}{$out_client}{-socket};
	    print $sock ":Server NOTICE :$message\n";
	}	       
    }
}

# Sends a message to a sock
#
sub sendSock {
    my($i_stat,$in_client,$message) = @_;
    foreach my $chan (keys %{$i_stat->{-channels}}) {
	print $in_client ":Server!bogus\@bogus NOTICE $chan :$message\n";
    }
    print $in_client ":Server NOTICE :$message\n";
}

# deletes a client from the list
#
sub delClient {
   my($i_stat,$client) = @_;
   $i_stat->{-sockstat}{-selhan}->remove($client);
   my $nick = $i_stat->{-clients}{$client}{-nick};
   delete $i_stat->{-clients}{$client};   
   massChanNotice($i_stat,"$nick has left the server :(\n");
}

# adds a client to the list
#
sub addClient {
    my($i_stat,$client) = @_;
    my($host,$ip) = returnHost($client);

    unless($ip) { # screwy DNS
	close($client);
	return;
    }

    $host ||= $ip;
    
    $i_stat->{-sockstat}{-selhan}->add($client);

    $i_stat->{-clients}{$client}{-socket} = $client;
    $i_stat->{-clients}{$client}{-host} = $host;
    $i_stat->{-clients}{$client}{-ip} = $ip;
    $i_stat->{-clients}{$client}{-init} = 0;
}

# initializes a new client to the existing irc connection - basically
#  just sends the initial server text and current channel info
#
sub initClient {
    my($i_stat,$client) = @_;
    $i_stat->{-clients}{$client}{-init} = 1;
    massChanNotice($i_stat,"$i_stat->{-clients}{$client}{-nick} has " .
		   "joined the server :)",$client);

    my($iref) = ( $i_stat->{-sockstat}{-irc_con_ref} );

    unless($$iref) {
	return;
    }    
    
    # first send initial 001 string

    print $client ":$i_stat->{-server} 001 $i_stat->{-me} " .
	":Welcome!\n";
    # should send some more interesting info

    # now send intro stuff we got earlier

    foreach my $line (@{$i_stat->{introtext}}) {
	print $client $line;
    }
    
    # now log on to all the channels - sending name list, current mode
    #  and maybe topic

    # not sending channel creation date, or topic set date

    foreach my $chan (keys %{$i_stat->{-channels}}) {

	# let client know that they have joined a channel
	my $send = ":$i_stat->{-me}!bogus\@bogus JOIN :$chan";
	print $client "$send\n";
		
	# current topic
	exists($i_stat->{-channels}{$chan}{-topic}) and
	    print $client ":$i_stat->{-server} 332 $i_stat->{-me} $chan :" .
		"$i_stat->{-channels}{$chan}{-topic}\n";

	# current mode sent on request

	# names
	my $i = 0;
	my @nicks = keys %{$i_stat->{-channels}{$chan}};
	my %trans = qw(o @ ov  @ vo @ v +); # mapping for ops and voices
	$trans{''} = ''; # blank mapping

        # weed out -limit, -key, etc and prepend + and @ to names
	@nicks = grep { ! /^-/ } @nicks; 
	@nicks = map { $trans{$i_stat->{-channels}{$chan}{$_}} . $_ } @nicks;

	my @group;
	while(@group = splice(@nicks,0,50)) { # send 1 message per 50 nicks
	    my $send = ":$i_stat->{-server} 353 $i_stat->{-me} @ $chan :" .
		join(' ',@group);
	    print $client "$send\n";
	}
	print $client ":$i_stat->{-server} 366 $i_stat->{-me} $chan :" .
	    "End of /NAMES list.\n"; 
    }
}

# connects to irc server, storing socket in the scalar ref it gets
#
sub conIrc {
    my($irc_con_ref,$address,$port) = @_;
    $port ||= 6667;   
    $$irc_con_ref = new IO::Socket::INET (
					  PeerPort => $port,
					  PeerAddr => $address,
					  Proto => 'tcp',
					  Timeout => 10
					  );
}

# returns hostname and ip of a socket, or nothing if there was an error
#
sub returnHost {
    my $client = shift;
    my $iaddr;
    (undef,$iaddr) = unpack_sockaddr_in(getpeername($client));
    my $ip = inet_ntoa($iaddr);
    my $hostname = gethostbyaddr($iaddr,AF_INET);
    my @name_lookup = gethostbyname($hostname);
    my $match = grep { $iaddr eq $_ } @name_lookup;
    if ($match) {
	return($hostname,$ip);
    } 
}

# daemonizes the process
#
sub daemonize {
    my $kid = fork(); # fork
    if ($kid) { # parent
        print "Process launched into background ($kid)\n";
        exit;
    } else { # daemonize!
	require POSIX;
        chdir("/") or die "No chdir: $!";
        umask(0);
        open(STDIN, "/dev/null") or die "No close STDIN: $!";
        open(STDOUT, "/dev/null") or die "No close STDOUT: $!";
        open(STDERR, "/dev/null") or die "No close STDERR: $!";
        POSIX::setsid();
    }
}

=head1 NAME

mird.pl - Multi-user Irc Relay Daemon

=head1 SYNOPSIS

B<perl> B<mird.pl> [ B<-f> ] [ B<-m maxcon> ] [ B<-p localport> ] [ B<-s server> ]

then

irc SomeNick serveraddress:port

BitchX SomeNick serveraddress:port

or your client of choice...

=head1 DESCRIPTION

B<mird.pl> is a irc server/client hybrid.  It accepts connections from irc clients - from the client's perspective, it is a normal irc session.  Meanwhile, it makes a connection out to some irc server.  The end result is that all client share the *same* connection to irc.  The first client to connect to the daemon supplies the NICK and USER strings.  Afterwards, the daemon maintains all irc state info so that future connections can hop right in.

=head1 OPTIONS

You can edit these defualts at the top of the script file.

=item B<-m maxcon> 

Maximum number of local connections

=item B<-p localport> 

Local port to listen for incoming connections

=item B<-s servername>

Default server - after the first client connects, the daemon will attempt to connect to this server first

=item B<-f>

Stay in foreground

=back

=head1 USAGE

These commands are executed by first connecting to the running server and using the irc client's /raw or /quote command to bypass normal client functions.

=item B</quote SERVER servername port>

Connects the daemon to an irc server at the specified port

=item B</quote BIGNICK newnick> 

changes the main server nickname

=item B</quote SHUTDOWN>

shuts it all down

=item B</quote STATUS>

gives a brief status report about who is connected and where they are from

=back

=head1 NOTES

The /nick command will change your apparent nickname that the other people connected see, but it does not change the irc connection's nickname - use /quote BIGNICK for that.

Text that begins with a * is private - that is, its never sent to the irc server, and only connected clients can see it.  Great for providing running commentary.

Text that goes to a channel from a client is sent back to the other clients as if it came from yourself, with a <nick> listed of the local nick who sent it - makes sense when you see it happening.

/msg'es to a nick appear to the other local clients as if it came *from* that nick - this makes all conversion appear in the same query window.

/dcc from a client appears on all the local connections, but the recipient can still get the file. /dcc to the client goes to all clients, and its a free-for-all as to who gets to accept the file or chat request.

The server handles some CTCP requests and server PING's in order to prevent general confusion.  One ramification of this is that with a few simple modifications, you have an effective bouncer that maintains a constant connection to irc.

For the most part, this system assumes friendly sockets - that is, it assumes that sockets won't block and freeze everything up.  Server and client input is buffered, while output is not.

To keep things simple, the interal channel list is maintained in lower case.

Whatever you do, don't use the /quote SERVER command to connect to the daemon itself!  It causes a nasty feedback loop.  On second though, try it once, its pretty funny.

This was not designed with security in mind - it was designed for a bunch of friends getting on irc and screwing around, which doesn't really necessitate any password or hostname authentication.

=head1 TO DO

If this were to be finished properly, it would need non-blocking sockets, full input and output buffers, a basic security system, and some kind of configuration file - but as it is, its still pretty fun.

=head1 SEE ALSO

http://www.irchelp.org for lots of IRC information

=head1 COPYRIGHT

Copyright 2000 Ben Kennedy.

This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 AUTHOR

Ben Kennedy <bkennedy@hmsonline.com> (Tikiman on EFnet)

=cut





