#$Id: Tftpd.pm,v 1.4 2000/04/03 06:47:34 ht8 Exp $

=head1 NAME

Tftpd - partial implementation of trivial ftp server

=head1 SYNOPSIS

=head1 DESCRIPTION

This implementation is not complete. The request to write (WRQ) and
the request for writing data (DATA) are NOT implemented. There is only
one client allowed to connect to the server at any time.

The main purpose to write this modules was to make it possible to
boot an lps17 (DEC Printserver 17) from NT 4.0 and W98 respectively,
as I found it inconvenient to start a window based TFTP server, a
window based bootp server and the script to download the initialization 
files. 

=head1 EXAMPLE

=cut


package Tftpd;


use strict;
use File::Basename;
use 5.004;
use Getopt::Std;
use Carp qw(carp);
use Errno;
#use IO::Socket::INET;
use IO::Socket;

=head1 METHODS

=head2 new (top_dir=>/tftpboot [,verbose=>0|1|2]  [,non_blocking=>0|1]))

=over 5

=item top_dir

Access will be restricted to files within given top directory. If the 
parameter is not specified, /tftpboot will be taken.

=item verbose

The following levels are currently supported:

0 no debug messages

1 give some info

2 trace entries into library methods

=item non_blocking

If true the socket will be put into non blocking mode.

=back

=cut 

sub new {
  my $class             = shift;
  my %pars              = @_;
  my $self              = {};
  my ($key, $val);
  my $errors            = 0;
  local $_;
  
  $self->{dbg}          = 0;
  $self->{non_blocking} = 0;
  $self->{connections}  = 0;
  $self->{request}{data}= ();
  $self->{response}{bnr}= -1;
  $self->{top_dir}      = "/tftpboot";
  
  
  while (($key, $val) = each %pars) {
    if    ($key =~ /^top_dir$/i) {
      $self->{top_dir} = $val;
      $self->{top_dir} =~ s:\\:/:g; #make sure MS Dos Files match
    }
    elsif ($key =~ /^non_blocking$/i) {
      $self->{non_blocking} = $val;
    }
    elsif ($key =~ /^verbose$/i) {
      $self->{dbg} = $val;
    }
    else {
      carp "Parameter \"$key\"=>\"$val\" is not supported";
      $errors += 1;
    }
  }
  return undef if $errors;
  
  unless ($self->{sock_tftp} = IO::Socket::INET->new
          (LocalPort=> 'tftp(69)',
           Proto    => 'udp')) {
    carp "Failed to establish tftp socket: $!";
    return undef;
  }
  
  
  if ($self->{non_blocking}){
    use Fcntl;
    
    my $flags = 0;
    fcntl($self->{sock_tftp}, F_GETFL, $flags)
      or die "Couldn't get flags for HANDLE : $!\n";
    $flags |= O_NONBLOCK;
    fcntl($self->{sock_tftp}, F_SETFL, $flags)
      or carp "Couldn't put tftp handle into non-blocking mode: $!";
  }
  
  bless $self, $class;
}


=head2 recv

If in blocking mode, the function will wait for a message to arrive.
If called in non-blocking mode, the function will return EAGAIN if there was
no data.

If there is a message, it will be examined and processed according to 
the given opcode.

returns: 

- undef if error occured. 

- "Xfer complete" if file transfer successfully completed


=cut
sub recv {
  my $self = shift;
  my $stat = $self->{sock_tftp}->recv($self->{request}{data}, 1024);

  &log("Received message") if $self->{dbg} > 1;
  if (!defined $stat) {
    return $!{EAGAIN} if $! == $!{EAGAIN};
    die "recv failed: $!";
  }
  return ($self->parse_msg);

}


sub parse_msg {
  my $self = shift;
  my @opcodes = qw (undefined RRQ WRQ DATA ACK ERROR);

  &log("entered parse_msg") if $self->{dbg} > 1;
  my ($x, $str);
  ($x, $self->{request}{opcode}, $str) = 
    unpack "CCa300", $self->{request}{data};

  printf "Message type is %s ($self->{request}{opcode})\n", 
    $opcodes[$self->{request}{opcode}] if $self->{dbg} > 1;

  if ($self->{request}{opcode} == 2 or #Write request (WRQ)
      $self->{request}{opcode} == 3) { #Data arrived  (DATA)
    carp "Write operations are not supported";
    $self->send_errmsg(4); # Illegal TFTP operation
    return undef;
  }
  elsif ($self->{request}{opcode} == 1) {
    &log("received TFTP Read Request (RRQ)") if $self->{dbg} > 0;
    ## Read request (RRQ)
    my ($filename, $mode);
    if ($str =~ m{([^\000]+)\000([^\000]+)\000$}) {
      ($filename, $mode) = ($1, uc $2);
      unless ($mode eq "NETASCII" or $mode eq "OCTET" or 
              $mode eq "MAIL") {
        my $msg = "Invalid mode \"$mode\"";
        carp $msg;
        $self->send_errmsg(0, $msg); 
        return undef;
      }

      $filename =~ s:\\:/:g; #make sure MS Dos Files match
      &log("I was asked to deliver \"$filename\"") if $self->{dbg} > 0;
      unless (-r $filename and $filename =~ m{^$self->{top_dir}}i) {
        my $msg = "cannot send file $filename, does not exist or no privilege";
        carp "$msg";
        carp "Hint: $filename is not within subtree $self->{top_dir}";
        $self->send_errmsg(0, $msg); 
        return undef;
      }

      if ($self->{connections} > 0) {
        my $msg = "tftpd is already serving \"$self->{request}{peer}\"";
        $self->send_errmsg(0, $msg); 
        carp "$msg $self->{connections}";
        return undef;
      }
      $self->{connections} += 1;

      my ($port, $ipaddr) = sockaddr_in($self->{sock_tftp}->peername);
      $self->{request}{peer} = gethostbyaddr($ipaddr, AF_INET);
      if (!defined $self->{request}{peer}) {
        #make sure we show at least an IP address
        $self->{request}{peer} = sprintf "%d.%d.%d.%d", unpack ("C4", $ipaddr);
      }
      &log(sprintf "\"%s\" is requesting \"%s\" in mode \"%s\"", 
           $self->{request}{peer}, $filename, $mode);

      unless (open FH, "$filename") {
        my $msg = "Failed to open $filename: $!";
        carp $msg;
        $self->send_errmsg(0, $msg); 
        return undef;
      }
      binmode FH if ($^O eq "MSWin32" and $mode eq "OCTET");
      $self->{response}{bnr}    =  1;
      $self->{response}{len}    = read FH,$self->{request}{current_block}, 512;
      $self->{response}{header} = pack "CCn", 0, 3, $self->{response}{bnr};
      my $stat = $self->{sock_tftp}->send($self->{response}{header} . 
                                          $self->{request}{current_block});
      if (!defined $stat) {
        carp "Unable to send data to peer: $!";
        return undef;
      }
      else {
        &log("Successfully sent first $stat bytes to peer") 
          if $self->{dbg} > 1;
      }
    }
    else {
      carp "received invalid TFTP Read Request (RRQ)";
      return undef;
    }
      }
  elsif ($self->{request}{opcode} == 4) {
    ## ACK
    my ($a, $b, $bnr) = unpack "CCn", $self->{request}{data};
    unless ($bnr == $self->{response}{bnr}) {
      my $msg = "invalid TFTP ACK received, actual: $bnr, ";
      $msg   .= "expected: $self->{response}{bnr}";
      carp $msg;
      $self->send_errmsg(0, $msg); # Unknown transfer ID
      return undef;
    }

    &log(sprintf "Received ACK %4d", $bnr) if $self->{dbg} > 1;
    ## Received valid ACK, we may send more data
    if ($self->{response}{len} < 512) {
      ## ACK of last block received, close file.
      $self->{connections} -= 1;
      close FH;
      if ($self->{dbg} > 0) {
        my $cnt = ($self->{response}{bnr} - 1) * 512 + $self->{response}{len};
        &log("$cnt bytes sent. TFTP RRQ successfully completed"); 
      }
      return "Xfer complete";
    }
    else {
      $self->{response}{bnr}++;
      $self->{response}{len} = read FH, $self->{request}{current_block}, 512;
      my $stat = $self->{sock_tftp}->send(pack "CCnA$self->{response}{len}",  
                                          0, 3, $self->{response}{bnr},
                                          $self->{request}{current_block});
      unless (defined $stat) {
        carp "Unable to send data to peer: $!";
        return undef;
      }
      else {
        &log("successfully sent pkg # $self->{response}{bnr} ($stat bytes)") 
          if $self->{dbg} > 1;
      }
    }
  }
  elsif ($self->{request}{opcode} == 5) {
    my ($errno, $msg);
    ($x, $errno, $msg) = unpack "CCa300", $str;
    &log("Error $errno received: \"$msg\"") if $self->{dbg} > 0;
    close FH;
    $self->{connections} = 0;
    return "Xfer complete";
  }  
  else {
   &log("unrecognised operation: $self->{request}{opcode}");
  }
}

=head2 send_errmsg(code [,error_message])

Code 1 to 6 are predefined errors. If any of these codes are passed to
send_errmsg, the parameter error_message will be ignored.

 1 File not found
 2 Access violation
 3 Disk full or allocation exceeded
 4 Illegal TFTP operation
 5 Unknown transfer ID
 6 No such user

If code is set to 0, error_message is sent to the peer.

=cut
sub send_errmsg {
  my $self = shift;
  my ($code, $txt) = @_;
  my @txts = (undef,                    'File not found',                    
              'Access violation',       'Disk full or allocation exceeded',  
              'Illegal TFTP operation', 'Unknown transfer ID',                 
              'No such user');

  unless (defined $code) {
    carp "must specify error code";
    return undef;
  }

  unless ($code <= 6 and $code >= 0) {
    carp "code out of range";
    return undef;
  }

  $txt = "no message given"       if (!defined $txt and $code == 0);
  $txt = $txts[$code]             if $code < 7 and $code > 0;
  &log("Error code: $code, text: \"$txt\"") if $self->{dbg} > 0;
  
  $self->{response}{header} = pack "C4", 0, 5, 0, $code;
  my $term = pack "C", 0;
  $self->{sock_tftp}->send($self->{response}{header} . $txt . $term);
}

=head2 disconnect (msg)


Use this routine to clean up internal state in case a connection
times out.

=cut

sub disconnect {
  my $self = shift;
  my $msg  = shift;

  close FH;
  $self->send_errmsg(0, $msg);
  $self->{connections} = 0;
}

sub log {
  my ($txt) = @_;
  my $date  = localtime;
  ($date)   = ($date =~ /^... (...\s+\d{1,2} ..:..:..)/);
  printf "$date: %s\n", $txt;
}

1;


=head1 AUTHOR

Beat.Hess@cyberdude.com

=cut
